;;;; Copyright  2006 Jeremy English <jhe@jeremyenglish.org>
;;;; 
;;;; Permission to use, copy, modify, distribute, and sell this software and its
;;;; documentation for any purpose is hereby granted without fee, provided that
;;;; the above copyright notice appear in all copies and that both that
;;;; copyright notice and this permission notice appear in supporting
;;;; documentation.  No representations are made about the suitability of this
;;;; software for any purpose.  It is provided "as is" without express or 
;;;; implied warranty.
;;;; 
;;;; Created: 12-July-2006 

(in-package music-theory)

(defun define-string (open-key frets)
		(labels ((get-string (key ls fret)
			   (if (> fret frets) ls			       
			       (get-string 
				(first (H key))
				(cons (H key) ls) 
				(+ fret 1)))))
		  (nreverse (get-string open-key (list (list open-key)) 1))))

(defun define-guitar (&key (tuning '(e a d g b e)) (frets 12))
  (nreverse (loop for i in tuning collecting (define-string i frets))))

(defun fret-note (guitar &key (fret 1) (string nil))
  (if (null string)
      (loop for i in guitar collecting (nth fret i))
      (nth fret (nth (- string 1) guitar))))

(defun fret-reachable (cf nf &key (threshold 5))
  (if (< (abs (- cf nf)) threshold) t nil))

(defun notes-on-string (guitar &key (string 1))
  (nth (- string 1) guitar))

(defun frets-for-note (guitar note &key (string 1))
  (labels ((get-fret (ls start)
	     (let ((fret 
		    (find-note note 
			       (notes-on-string guitar :string string) 
			       start)))
	       (if fret 
		   (get-fret (cons fret ls) (+ fret 1))
		   (nreverse ls)))))
    (get-fret '() 0)))


(defun inversion (chord)
  (append (cdr chord) (list (car chord))))

(defun good-fret (old-fret new-fret fret-list &key (threshold 5))
  (if old-fret
      (if new-fret
	  (notany #'null (mapcar #'(lambda (x) (if x (fret-reachable x new-fret :threshold threshold) t)) fret-list))
	  nil)
      t))

(defun find-chord (guitar chord &key (root-string nil root-string-supplied-p) (threshold 5))
  "Finds a fingering for the givin chord with the root on the
root-string.  Sometimes it dosen't know when you run out of
fingers. If you are flexibale you may want to adjust the threshold to
get different voicings.  Threshold controls the maximum distance
between to frets."
  (labels ((get-frets (string-num chord old-fret fret-list root-string try)
	     (if (> string-num 0)
		 (let ((fret (if (and chord (<= string-num root-string))
				 (first 
				  (frets-for-note guitar 
						  (first chord) 
						  :string string-num))
				 nil)))
		   (if (or (> try (length chord)) (good-fret old-fret fret fret-list :threshold threshold))
		       (get-frets (1- string-num) 
				  (if fret (inversion chord) chord)
				  (if fret fret old-fret)
				  (if (good-fret old-fret fret fret-list :threshold threshold)
				      (cons fret fret-list)
				      (cons nil fret-list))
				  root-string
				  0)
		       (get-frets string-num
				  (inversion chord)
				  old-fret
				  fret-list
				  root-string
				  (1+ try))))
		 (nreverse fret-list))))
    (get-frets (length guitar) chord nil (list) 
	       (if root-string-supplied-p
		   root-string
		   (length guitar))
	       0)))

(defun print-fret-board (guitar)
  (format t "~{String ~{ ~a: ~{ ~2a ~}~}~%~}" 
	  (loop for i in guitar 
		for j from 1
		collecting (list j (mapcar #'car i)))))

(defun notes-at-fret (guitar fret &key (string nil))
  (if string
      (nth fret (notes-on-string guitar :string string))
    (loop for i in guitar collecting (nth fret i))))

(defun note-guessing-game (guitar &key (fret nil) (string nil) (escape 'quit))
  (let ((n-strings (length guitar))
	(n-frets (length (first (define-guitar)))))
    (loop
     (let* ((fret (if fret fret (random n-frets)))
	    (string (if string string (1+ (random n-strings))))
	    (note (notes-at-fret guitar fret :string string)))	      
       (format t "Note on fret ~a string ~a~%" fret string)
       (let ((ans (read)))
	 (cond
	  ((equal ans escape) (return))
	  ((position ans note) (format t "Right!~%"))
	  (t
	   (format t "Wrong!~%    the note on fret ~a string ~a is ~a ~%" fret string note))))))))